package net.cnri.apps.doserver;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.Iterator;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.cnri.apps.doserver.HTTPConnectionHandler.HTTPRepositoryInterface;
import net.cnri.do_api.DataElement;
import net.cnri.do_api.DigitalObject;
import net.cnri.do_api.Repository;
import net.cnri.do_api.SearchResult;
import net.cnri.do_api.Repository.CloseableIterator;
import net.cnri.dobj.DOAuthentication;
import net.cnri.dobj.DOConstants;
import net.cnri.dobj.DOException;
import net.cnri.dobj.HeaderItem;
import net.cnri.dobj.HeaderSet;
import net.cnri.util.StringUtils;
import net.handle.hdllib.Util;

public class RESTServlet extends HttpServlet {
    public static final int BUFFER_SIZE = 1024;
    public static final String XML = "application/xml";
    private static final String PLAINTEXT = "text/plain";

    private final Main serverMain;
    private final HTTPConnectionHandler httpConnectionHandler;
    
    public RESTServlet(Main serverMain, HTTPConnectionHandler httpConnectionHandler) {
        this.serverMain = serverMain;
        this.httpConnectionHandler = httpConnectionHandler;
    }
    
    @Override
    protected void service(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        try {
            super.service(request,response);
        }
        catch(DOException e) {
            if(e.getErrorCode()==DOException.PERMISSION_DENIED_ERROR) {
                if(!response.isCommitted()) {
                    response.reset();
                    if(request.getAttribute(HTTPConnectionHandler.DO_AUTHENTICATION_REQ_ATTRIBUTE)==null) {
                        response.setHeader("WWW-Authenticate", "Basic realm=\"hdl:" + serverMain.getServerID() + "\"");
                        response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
                    }
                    else {
                        response.setStatus(HttpServletResponse.SC_FORBIDDEN);
                    }
                }
                else {
                    System.err.println("REST server: Permission error after response committed:");
                    e.printStackTrace();
                }
            }
            else {
                if(!response.isCommitted()) {
                    response.reset();
                    response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
                    response.setContentType("text/plain");
                    response.setCharacterEncoding("UTF-8");
                    if(!"HEAD".equals(request.getMethod())) e.printStackTrace(response.getWriter());
                }
                e.printStackTrace();
            }
        }
        catch(Exception e) {
            if(!response.isCommitted()) {
                response.reset();
                response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
                response.setContentType("text/plain");
                response.setCharacterEncoding("UTF-8");
                if(!"HEAD".equals(request.getMethod())) e.printStackTrace(response.getWriter());
            }
            e.printStackTrace();
        }
    }
    
    private static void writeDataElement(Writer w, String elementId, HeaderSet atts) throws IOException {
        w.write("        <el id=\"");
        w.write(xmlAttrEscape(elementId));
        if(atts==null) w.write("\"/>\n");
        else {
            w.write("\">\n");
            for(Iterator it=atts.iterator(); it.hasNext(); ) {
                HeaderItem attItem = (HeaderItem)it.next();
                w.write("            <att name=\"");
                w.write(xmlAttrEscape(attItem.getName()));
                w.write("\" value=\"");
                w.write(xmlAttrEscape(attItem.getValue()));
                w.write("\" />\n");
            }
            w.write("        </el>\n");
        }
    }
    
    private static void writeXML(HttpServletResponse response, DigitalObject dobj, DataElement element, boolean listAtt, boolean listElements) throws IOException {
        response.setContentType(XML);
        response.setCharacterEncoding("UTF-8");
        Writer w = response.getWriter();
        w.write("<dor id=\"");
        w.write(xmlAttrEscape(dobj.getRepository().getID()));
        w.write("\">\n");
        w.write("    <do id=\"");
        w.write(xmlAttrEscape(dobj.getID()));
        w.write("\">\n");
        if(!listElements && element==null) {
            HeaderSet atts = dobj.getAttributes();
            for(Iterator it=atts.iterator(); it.hasNext(); ) {
                HeaderItem attItem = (HeaderItem)it.next();
                w.write("        <att name=\"");
                w.write(xmlAttrEscape(attItem.getName()));
                w.write("\" value=\"");
                w.write(xmlAttrEscape(attItem.getValue()));
                w.write("\" />\n");
            }
        }
        if(element!=null || !listAtt) {
            if(element==null) {
                String elementIDs[] = dobj.listDataElements();
                for(int i=0; elementIDs!=null && i<elementIDs.length; i++) {
                    HeaderSet atts = null;
                    if(!listElements) {
                        atts = dobj.getDataElement(elementIDs[i]).getAttributes();
                    }
                    writeDataElement(w,elementIDs[i],atts);
                }
            }
            else {
                HeaderSet atts = element.getAttributes();
                writeDataElement(w,element.getDataElementID(),atts);
            }
        }
        w.write("    </do>\n");
        w.write("</dor>\n");
    }
    
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        DOAuthentication auth = httpConnectionHandler.getAuthentication(request,response);
        Repository repo = new HTTPRepositoryInterface(serverMain, request, auth.getID());
        ParsedURL url = new ParsedURL(request);

        // LIST ELEMENTS OR SEARCH
        if(url.doId==null) {
            if(request.getQueryString()!=null && !request.getQueryString().isEmpty()) {
                handleRestSearch(request,response,repo);
                return;
            }
            
            // Return XML Document Listing all DOs accessible through this DOR; if none, return empty list
            response.setStatus(HttpServletResponse.SC_OK);
            response.setContentType(XML);
            response.setCharacterEncoding("UTF-8");
            Iterator allObj = repo.listObjects();
            Writer w = response.getWriter();
            w.write("<dor id=\"");
            w.write(xmlAttrEscape(repo.getID()));
            w.write("\">\n");
            while(allObj.hasNext()) {
                String objID = String.valueOf(allObj.next());
                w.write("    <do id=\"");
                w.write(xmlAttrEscape(objID));
                w.write("\" />\n");
            }
            w.write("</dor>\n");
            return;
        }
        else {
            if(request.getQueryString()!=null) {
                response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                return;
            }

            if(!repo.verifyDigitalObject(url.doId)) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                return;
            }
            DigitalObject dobj = repo.getDigitalObject(url.doId);
            
            DataElement element = null;
            if(url.elId!=null) {
                if(!dobj.verifyDataElement(url.elId)) {
                    response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                    return;
                }
                element = dobj.getDataElement(url.elId);
            }
            
            // RETRIEVE AN ATTRIBUTE
            if(url.attName!=null) {
                String value;
                if(element==null) value = dobj.getAttribute(url.attName,null);
                else value = element.getAttribute(url.attName,null);
                
                if(value==null) {
                    response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                    return;
                }
                else {
                    response.setStatus(HttpServletResponse.SC_OK);
                    response.setContentType(PLAINTEXT);
                    response.setCharacterEncoding("UTF-8");
                    response.getWriter().write(value);
                }
            }
            // RETRIEVE AN ELEMENT'S DATA
            else if(element!=null && !url.endsAtt) {
                String mimetype = element.getAttribute(DOConstants.MIME_TYPE_ATTRIBUTE,null);
                if(mimetype!=null) response.setContentType(mimetype);
                String filename = element.getAttribute(DOConstants.FILE_NAME_ATTRIBUTE, url.elId);
                response.setHeader("Content-Disposition", "attachment; filename=\"" + escapeFilename(filename) + "\"");
                response.setHeader("Content-Length", String.valueOf(element.getSize()));
                
                InputStream in = element.read();
                OutputStream out = response.getOutputStream();
                byte buf[] = new byte[BUFFER_SIZE];
                int r;
                while((r=in.read(buf))>=0) out.write(buf, 0, r);
                out.close();
                return;
            }
            // RETRIEVE (SOME PART OF) THE DO'S STRUCTURE
            else {
                response.setStatus(HttpServletResponse.SC_OK);
                writeXML(response,dobj,element,url.endsAtt,url.endsEl);
            }
        }
    }

    @Override
    protected void doDelete(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        if(request.getQueryString()!=null) {
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
        ParsedURL url = new ParsedURL(request);
        if(url.doId==null) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow","GET, HEAD, POST");
            return;
        }
        if(url.endsAtt) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow","GET, HEAD");
            return;
        }
        if(url.endsEl) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow", "GET, HEAD");
            return;
        }

        DOAuthentication auth = httpConnectionHandler.getAuthentication(request,response);
        Repository repo = new HTTPRepositoryInterface(serverMain, request, auth.getID());
        
        if(!repo.verifyDigitalObject(url.doId)) {
            response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            return;
        }
        if(url.elId==null && url.attName==null) {
            repo.deleteDigitalObject(url.doId);
            response.setStatus(HttpServletResponse.SC_NO_CONTENT);
            return;
        }
        DigitalObject dobj = repo.getDigitalObject(url.doId);

        if(url.elId!=null) {
            if(!dobj.verifyDataElement(url.elId)) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                return;
            }

            if(url.attName==null) {
                dobj.deleteDataElement(url.elId);
                response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                return;
            }
            
            dobj.getDataElement(url.elId).deleteAttribute(url.attName);
        }
        else {
            dobj.deleteAttribute(url.attName);
        }
        response.setStatus(HttpServletResponse.SC_NO_CONTENT);
    }
    
    @Override
    protected void doPut(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        doPost(request,response);
    }
    
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        if(request.getQueryString()!=null) {
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
        if(request.getHeader("Content-Encoding")!=null || request.getHeader("Content-Range")!=null) {
            response.setStatus(HttpServletResponse.SC_NOT_IMPLEMENTED);
            return;
        }

        ParsedURL url = new ParsedURL(request);
        if(url.doId==null && request.getMethod().equals("PUT")) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow","GET, HEAD, POST");
            return;
        }
        if(url.endsAtt) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow","GET, HEAD");
            return;
        }
        if(url.endsEl) {
            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
            response.setHeader("Allow", "GET, HEAD");
            return;
        }

        DOAuthentication auth = httpConnectionHandler.getAuthentication(request,response);
        Repository repo = new HTTPRepositoryInterface(serverMain, request, auth.getID());
    
        // CREATE A NEW OBJECT
        if(url.elId==null && url.attName==null) {
            /* Creates a new DO with the speciﬁed handle. If created, the DOR will assign
             * assign the speciﬁed handle to the DO and return it in a response with 
             * HTTP 201 Created. 
             */
            if(request.getInputStream()!=null && request.getInputStream().read()!=-1) {
                response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                return;
            }

            if(url.doId!=null && repo.verifyDigitalObject(url.doId)) {
                response.setStatus(HttpServletResponse.SC_CONFLICT);
                return;
            }
            else {
                DigitalObject dobj = repo.createDigitalObject(url.doId);
                response.setStatus(HttpServletResponse.SC_CREATED);
                String uri = "/do/" + StringUtils.encodeURLComponent(dobj.getID());
                response.setHeader("Location",uri);
                writeXML(response,dobj,null,false,false);
                return;
            }
        }
        else {
            if(!repo.verifyDigitalObject(url.doId)) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                return;
            }
            DigitalObject dobj = repo.getDigitalObject(url.doId);

            // PUTTING AN ATTRIBUTE
            if(url.attName!=null) {
                String value = streamToString(request.getInputStream(),request.getCharacterEncoding());

                DataElement element = null;
                if(url.elId!=null) {
                    boolean exists = dobj.verifyDataElement(url.elId);
                    element = dobj.getDataElement(url.elId);
                    if(!exists) {
                        element.write(new ByteArrayInputStream(new byte[0]));
                    }
                }

                boolean attributeExisted;
                if(element==null) {
                    attributeExisted = dobj.getAttribute(url.attName, null)!=null;
                    // Replaces the plaintext value of this attribute. If created,returns HTTP 201 Created. 
                    dobj.setAttribute(url.attName, value);
                }
                else {
                    attributeExisted = element.getAttribute(url.attName, null)!=null;
                    // Replaces the plaintext value of this attribute. If created,returns HTTP 201 Created. 
                    element.setAttribute(url.attName, value);
                }

                if(!attributeExisted) {
                    response.setStatus(HttpServletResponse.SC_CREATED);
                    String path = "/do/" + StringUtils.encodeURLComponent(url.doId);
                    if(element!=null) path += "/el/" + StringUtils.encodeURLComponent(url.elId);
                    path += "/att/" + StringUtils.encodeURLComponent(url.attName);
                    response.setHeader("Location",path);
                    return;
                }
                else {
                    response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                    return;
                }
            } 
            else {
                // PUTTING AN ELEMENT'S DATA
                
                boolean exists = dobj.verifyDataElement(url.elId);
                DataElement element = dobj.getDataElement(url.elId);
                InputStream in = request.getInputStream();
                if(in!=null) {
                    element.write(in);
                    String mimetype = request.getContentType();
                    if(mimetype!=null) element.setAttribute(DOConstants.MIME_TYPE_ATTRIBUTE,mimetype);
                }
                else if(!exists) {
                    element.write(new ByteArrayInputStream(new byte[0]));
                }
                if(!exists) {
                    response.setStatus(HttpServletResponse.SC_CREATED);
                    response.setHeader("Location","/do/" + StringUtils.encodeURLComponent(url.doId) + "/el/" + StringUtils.encodeURLComponent(url.elId));
                    return;
                }
                else {
                    response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                    return;
                }
            }
        }
        
    }
    
    private void handleRestSearch(HttpServletRequest request, HttpServletResponse response, Repository repo) throws IOException {
        String query = request.getParameter("query");
        if(query==null) {
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
        
        Writer w;
        CloseableIterator<HeaderSet> iter = repo.search(query);
        try {
            response.setContentType(XML);
            response.setCharacterEncoding("UTF-8");
            w = response.getWriter();
            w.write("<dor id=\"");
            w.write(xmlAttrEscape(repo.getID()));
            w.write("\">\n");
            w.write("    <query>");
            w.write(StringUtils.cgiEscape(query));
            w.write("</query>\n");
            while(iter.hasNext()) {
                HeaderSet headerSet = iter.next();
                SearchResult searchResult = SearchResult.ofHeaderSet(headerSet);
                w.write("    <do");
                if(searchResult.getObjectID()!=null) {
                    w.write(" id=\"");
                    w.write(xmlAttrEscape(searchResult.getObjectID()));
                    w.write("\"");
                }
                if(!repo.getID().equals(searchResult.getRepoID())) {
                    w.write(" repo=\"");
                    w.write(xmlAttrEscape(searchResult.getObjectID()));
                    w.write("\"");
                }
                if(searchResult.getScore()!=null) {
                    w.write(" score=\"");
                    w.write(xmlAttrEscape(searchResult.getScore()));
                    w.write("\"");
                }
                w.write(">\n");

                if(searchResult.getFields()!=null) {
                    for(Map.Entry<String,String> entry : searchResult.getFields().entrySet()) {
                        w.write("        <field name=\"");
                        w.write(xmlAttrEscape(entry.getKey()));
                        w.write("\" value=\"");
                        w.write(xmlAttrEscape(entry.getValue()));
                        w.write("\" />\n");
                    }
                }
                if(searchResult.getAtts()!=null) {
                    for(Map.Entry<String,String> entry : searchResult.getAtts().entrySet()) {
                        w.write("        <att name=\"");
                        w.write(xmlAttrEscape(entry.getKey()));
                        w.write("\" value=\"");
                        w.write(xmlAttrEscape(entry.getValue()));
                        w.write("\" />\n");
                    }
                }
                
                if(searchResult.getElements()!=null) {
                    for(SearchResult.Element element : searchResult.getElements().values()) {
                        w.write("        <el id=\"");
                        w.write(xmlAttrEscape(element.getElementID()));
                        w.write("\">\n");
                     
                        if(element.getFields()!=null) {
                            for(Map.Entry<String,String> entry : element.getFields().entrySet()) {
                                w.write("            <field name=\"");
                                w.write(xmlAttrEscape(entry.getKey()));
                                w.write("\" value=\"");
                                w.write(xmlAttrEscape(entry.getValue()));
                                w.write("\" />\n");
                            }
                        }
                        if(element.getAtts()!=null) {
                            for(Map.Entry<String,String> entry : element.getAtts().entrySet()) {
                                w.write("            <att name=\"");
                                w.write(xmlAttrEscape(entry.getKey()));
                                w.write("\" value=\"");
                                w.write(xmlAttrEscape(entry.getValue()));
                                w.write("\" />\n");
                            }
                        }
                        if(element.getPAtts()!=null) {
                            for(Map.Entry<String,String> entry : element.getPAtts().entrySet()) {
                                w.write("            <patt name=\"");
                                w.write(xmlAttrEscape(entry.getKey()));
                                w.write("\" value=\"");
                                w.write(xmlAttrEscape(entry.getValue()));
                                w.write("\" />\n");
                            }
                        }
                        if(element.getMPAtts()!=null) {
                            for(Map.Entry<String,String> entry : element.getMPAtts().entrySet()) {
                                w.write("            <mpatt name=\"");
                                w.write(xmlAttrEscape(entry.getKey()));
                                w.write("\" value=\"");
                                w.write(xmlAttrEscape(entry.getValue()));
                                w.write("\" />\n");
                            }
                        }

                        if(element.getText()!=null) {
                            w.write("            <text>");
                            w.write(StringUtils.cgiEscape(element.getText()));
                            w.write("</text>\n");
                        }

                        w.write("        </el>\n");
                    }
                }
                
                w.write("    </do>\n");
            }
        }
        finally {
            iter.close();
        }
        w.write("</dor>\n");
    }
    

    static String urlDecode(String s) {
        return StringUtils.decodeURLIgnorePlus(s);
    }
    
    //Read from stream and return string
    static String streamToString(InputStream input, String encoding) throws IOException{
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        byte buf[] = new byte[BUFFER_SIZE];
        int r;
        while((r=input.read(buf))>=0)
            bout.write(buf, 0, r);
        if(encoding==null) return Util.decodeString(bout.toByteArray());
        else {
            return new String(bout.toByteArray(), encoding);
        }
    }

    static String escapeFilename(String filename) {
      // This is gruesome, but as of mid-2010 is the only reasonably portable solution
      try {
        String res = new String(filename.getBytes("ISO-8859-1"),"ISO-8859-1");
        return res.replace("\n","?").replace("\\","\\\\").replace("\"","\\\"");
      } catch(UnsupportedEncodingException e) {
        throw new AssertionError(e);
      }
    }

    static String xmlAttrEscape(String str) {
        if(str==null) return "null";
        StringBuilder sb = new StringBuilder("");
        for(int i=0;i<str.length();i++) {
          char ch = str.charAt(i);
          if(ch=='<')
            sb.append("&lt;");
          else if(ch=='>')
            sb.append("&gt;");
          else if(ch=='"')
            sb.append("&quot;");
          else if(ch=='&')
            sb.append("&amp;");
          else if(ch=='\'') 
            sb.append("&#39;");
          else if(ch=='\t') 
            sb.append("&#x9;");
          else if(ch=='\n') 
              sb.append("&#xA;");
          else if(ch=='\r') 
              sb.append("&#xD;");
          else
            sb.append(ch);
        }
        return sb.toString();
      }

    
    private static class ParsedURL {
        String doId;
        boolean endsEl;
        String elId;
        boolean endsAtt;
        String attName;
        
        ParsedURL(String path) {
            if(!path.endsWith("/")) path = path + "/";
            if(path.equals("/do/")) return;
            int elIndex = path.indexOf("/el/");
            int attIndex = path.indexOf("/att/");
            
            // avoid /do/.../att/.../el/...
            if(attIndex >= 0 && elIndex > attIndex) elIndex = -1;
            
            // look for terminal /el/ or /att/
            if(elIndex+4 == path.length()) {
                endsEl = true;
            }
            if(attIndex+5 == path.length()) {
                endsAtt = true;
            }
            
            if(elIndex >= 0) {
                doId = path.substring(4,elIndex);
                if(attIndex >= 0) {
                    elId = path.substring(elIndex + 4, attIndex);
                    if(!endsAtt) {
                        attName = path.substring(attIndex + 5, path.length()-1);
                    }
                }
                else {
                    if(!endsEl) {
                        elId = path.substring(elIndex + 4, path.length()-1);
                    }
                }
            }
            else {
                if(attIndex >= 0) {
                    doId = path.substring(4,attIndex);
                    if(!endsAtt) {
                        attName = path.substring(attIndex + 5, path.length()-1);
                    }
                }
                else {
                    doId = path.substring(4,path.length()-1);
                }
            }
            
            if(doId!=null) doId = urlDecode(doId);
            if(elId!=null) elId = urlDecode(elId);
            if(attName!=null) attName = urlDecode(attName);
        }
        
        ParsedURL(HttpServletRequest req) {
            this(pathOfReq(req));
        }

        static String pathOfReq(HttpServletRequest req) {
            String path = req.getServletPath();
            String pathInfo = req.getPathInfo();
            if(pathInfo!=null) path = path + pathInfo;
            return path;
        }
    }
}
