/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.apps.doutilgui;

import net.cnri.dobj.*;
import net.handle.hdllib.*;
import net.handle.awt.*;
import net.cnri.util.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.security.*;
import java.io.*;

/**
 * Window that can be used to gather authentication information from a user
 * such as their identifier and private/secret key.
 */
public class AuthWindow 
  extends JDialog 
  implements ActionListener
{
  private JComboBox authTypeChoice;
  
  private JTextField idField;
  private JPasswordField passField;
  private JLabel privKeyLabel;
  private JLabel keyFieldLabel;
  private JLabel passFieldLabel;
  private JButton anonButton;
  private JButton okButton;
  private JButton cancelButton;
  private JButton loadKeyButton;
  private String authRole = "default";
  private ActionListener extraHandler = null;
  
  private File privKeyFile = null;
  private PrivateKey privKey = null;
  private boolean wasCanceled = true;
  private StreamTable storedSettings = null;

  public AuthWindow(Frame owner)
    throws HeadlessException
  {
    this(owner, "default");
  }
  
  /** 
   * Constructor for a window object that provides an interface for a user
   * to enter their authentication information.  The given roleID provides
   * a key to the type of authentication that is being performed, so that
   * the appropriate default login and key file are used.
   */
  public AuthWindow(Frame owner, String roleID)
    throws HeadlessException
  {
    super(owner, "Authentication Details", true);

    if(roleID!=null)
      authRole = roleID;
    
    storedSettings = new StreamTable();
    
    try {
      storedSettings.readFromFile(System.getProperty("user.home", "")+
        File.separator + ".do_authentication");
    } catch (Exception e) {}

    String privKeyFileStr = storedSettings.getStr("keyfile_"+authRole, null);
    if(privKeyFileStr!=null) 
      privKeyFile = new File(privKeyFileStr);
    
    idField = new JTextField(storedSettings.getStr("id_"+authRole, ""), 30);
    passField = new JPasswordField(20);
    okButton = new JButton("OK");
    cancelButton = new JButton("Cancel");
    loadKeyButton = new JButton("Select...");
    
    anonButton = new JButton("Anonymous");
    authTypeChoice = new JComboBox(new String [] {
      "Private Key", "Password"
    });
    if(storedSettings.getStr("authtype", "").equals("seckey")) {
      authTypeChoice.setSelectedIndex(1);
    } else {
      authTypeChoice.setSelectedIndex(0);
    }
    
    privKeyLabel = new JLabel("no key loaded");
    privKeyLabel.setForeground(Color.gray);

    JPanel p = new JPanel(new GridBagLayout());
    p.setBorder(new javax.swing.border.EmptyBorder(10,10,10,10));
    int y = 0;
    p.add(new JLabel("Your ID: ", SwingConstants.RIGHT), 
          AwtUtil.getConstraints(0,y,0,0,1,1,new Insets(4,4,4,0),true,false));
    p.add(idField,
          AwtUtil.getConstraints(1,y++,1,0,2,1,new Insets(4,0,4,4),true,false));
    
    p.add(new JLabel("Authentication Type: ", SwingConstants.RIGHT),
          AwtUtil.getConstraints(0,y,0,0,1,1,new Insets(4,4,4,0),true,false));
    p.add(authTypeChoice,
          AwtUtil.getConstraints(1,y++,1,0,2,1,new Insets(4,0,4,4),true,false));
          
    
    p.add(keyFieldLabel = new JLabel("Private Key File: ", SwingConstants.RIGHT), 
          AwtUtil.getConstraints(0,y,0,0,1,1,new Insets(4,4,4,0),true,false));
    p.add(privKeyLabel,
          AwtUtil.getConstraints(1,y,1,0,1,1,new Insets(4,0,4,4),true,false));
    p.add(loadKeyButton,
          AwtUtil.getConstraints(2,y++,0,0,1,1,new Insets(4,0,4,4),true,false));
    
    p.add(passFieldLabel = new JLabel("Password: ", SwingConstants.RIGHT), 
          AwtUtil.getConstraints(0,y,0,0,1,1,new Insets(4,4,4,0),true,false));
    p.add(passField,
          AwtUtil.getConstraints(1,y++,1,0,2,1,new Insets(4,0,4,4),true,false));
    p.add(Box.createVerticalStrut(10),
          AwtUtil.getConstraints(1,y++,0,1,1,1,false,false));
    JPanel bp = new JPanel(new GridBagLayout());
    bp.add(anonButton,
           AwtUtil.getConstraints(0,0,0,0,1,1,false,false));
    bp.add(Box.createHorizontalStrut(20),
           AwtUtil.getConstraints(1,0,1,0,1,1,false,false));
    bp.add(cancelButton,
           AwtUtil.getConstraints(2,0,0,0,1,1,new Insets(0,0,0,20),false,false));
    bp.add(okButton,
           AwtUtil.getConstraints(3,0,0,0,1,1,false,false));
    p.add(bp,
          AwtUtil.getConstraints(0,y++,1,0,3,1,true,true));
    
    okButton.addActionListener(this);
    cancelButton.addActionListener(this);
    loadKeyButton.addActionListener(this);
    anonButton.addActionListener(this);
    authTypeChoice.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent evt) {
        authTypeSelected();
      }
    });
    
    getContentPane().add(p);
    getRootPane().setDefaultButton(okButton);
    
    //setRootPane().setDefaultButton(okButton);
    pack();
    setSize(getPreferredSize());
    setLocationRelativeTo(null);
    
    authTypeSelected();
  }
  
  
  public void setExtraHandler(String actionName, ActionListener handler) {
    if(actionName==null || handler==null)
      throw new NullPointerException();
    anonButton.setText(actionName);
    this.extraHandler = handler;
  }
  
  private void authTypeSelected() {
    boolean pubkey = authTypeChoice.getSelectedIndex()==0;
    keyFieldLabel.setVisible(pubkey);
    privKeyLabel.setVisible(pubkey);
    loadKeyButton.setVisible(pubkey);
    passFieldLabel.setVisible(!pubkey);
    passField.setVisible(!pubkey);
    if(pubkey) {
      updateKeyLabel();
    }
    validate();
  }
  
  private void updateKeyLabel() {
    PrivateKey key = privKey;
    if(key!=null) {
      byte encoded[] = key.getEncoded();
      String encodedStr = "???";
      if(encoded!=null)
        encodedStr = Util.decodeHexString(encoded, false);
      privKeyLabel.setText("Algorithm: "+key.getAlgorithm()+
          "; Format: "+key.getFormat());
    } else if(privKeyFile!=null) {
      privKeyLabel.setText(privKeyFile.getName());
    } else {
      privKeyLabel.setText("no key loaded - press the \"Load\" button");
    }
  }

  private void chooseKey() {
    JFileChooser chooser = new JFileChooser();
    chooser.setDialogTitle("Select your private key file");
    int returnVal = chooser.showOpenDialog(null);
    if(returnVal != JFileChooser.APPROVE_OPTION) {
      updateKeyLabel();
      return;
    }
    
    File f = chooser.getSelectedFile();
    if(f!=null && (privKeyFile==null || !f.equals(privKeyFile))) {
      // the user selected a different file
      privKey = null;
      privKeyFile = f;
    }
    updateKeyLabel();
    return;
  }
  
  private PrivateKey loadPrivateKeyFromFile() {
    if(privKey!=null) return privKey;
    
    if(privKeyFile!=null) {
      try {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        FileInputStream fin = new FileInputStream(privKeyFile);
        byte buf[] = new byte[256];
        int r = 0;
        while((r=fin.read(buf))>=0)
          bout.write(buf, 0, r);
        buf = bout.toByteArray();
        byte passphrase[] = null;
        if(Util.requiresSecretKey(buf)) {
          // ask the user for their secret key...
          PassphrasePanel pp = new PassphrasePanel();
          int result = 
            JOptionPane.showConfirmDialog(this, pp, "Enter Passphrase", 
                                          JOptionPane.OK_CANCEL_OPTION);
          if(result==JOptionPane.OK_OPTION) {
            passphrase = new String(pp.getPassphrase()).getBytes("UTF8");
          } else {
            return null;
          }
        }
        
        buf = Util.decrypt(buf, passphrase);
        
        return Util.getPrivateKeyFromBytes(buf, 0);
      } catch (Exception e) {
        JOptionPane.showMessageDialog(this, "Error: "+e, 
            "Error", JOptionPane.ERROR_MESSAGE);
      }
    } else {
      JOptionPane.showMessageDialog(this, "Error: Please specify a private key file", 
                                    "Error", JOptionPane.ERROR_MESSAGE);
    }
    return null;
  }
  
  private boolean checkInputs() {
    String id = idField.getText().trim();
    if(id==null || id.trim().length()<=0) {
      JOptionPane.showMessageDialog(this, 
                                    "Please enter your identifier",
                                    "No identifier specified",
                                    JOptionPane.ERROR_MESSAGE);
      return false;
    }
    
    boolean pubkey = authTypeChoice.getSelectedIndex()==0;
    if(pubkey) {
      if(privKey==null) {
        privKey = loadPrivateKeyFromFile();
        if(privKey==null) {
          return false;
        }
      }
    } else {
      // no way to screw up a password entry... it could be blank
    }
    return true;
  }
  
  private void storeValues() {
    storedSettings.put("id_"+authRole, idField.getText().trim());
    storedSettings.put("authtype", authTypeChoice.getSelectedIndex()==0 ? "pubkey" : "seckey");
    try {
      storedSettings.put("keyfile_"+authRole, privKeyFile.getCanonicalPath());
    } catch (Exception e) {}
    try {
      storedSettings.writeToFile(System.getProperty("user.home", "")+
        File.separator + ".do_authentication");
    } catch (Exception e) {}
  }
  
  /** Set the public key authentication information for this window */
  public void setAuthentication(String id, PrivateKey newKey) {
    idField.setText(id);
    this.privKey = newKey;
    updateKeyLabel();
  }
  
  /** Set the secret key authentication information for this window */
  public void setSecretKeyAuthentication(String id, String secretKey) {
    idField.setText(id);
    passField.setText(secretKey);
  }
  
  public boolean wasCanceled() {
    return wasCanceled;
  }
  
  
  public String getID() {
    return idField.getText().trim();
  }
  
  
  public PrivateKey getPrivateKey() {
    return privKey;
  }
  
  public String getSecretKey() {
    return new String(passField.getPassword());
  }
  
  
  public DOAuthentication getAuthentication() {
    if(authTypeChoice.getSelectedIndex()==0) {
      return new PKAuthentication(getID(), privKey);
    } else {
      return new SecretKeyAuthentication(getID(), Util.encodeString(new String(passField.getPassword())));
    }
  }
  
  
  public void actionPerformed(ActionEvent evt) {
    Object src = evt.getSource();
    if(src==okButton) {
      if(checkInputs()) {
        wasCanceled = false;
        setVisible(false);
        storeValues();
      }
    } else if(src==cancelButton) {
      wasCanceled = true;
      setVisible(false);
    } else if(src==anonButton) {
      if(extraHandler!=null) {
        extraHandler.actionPerformed(evt);
      } else {
        try {
          InputStream pkIn = 
          PKAuthentication.class.getResourceAsStream("/net/cnri/dobj/etc/anonymous_privkey.bin");
          ByteArrayOutputStream bout = new ByteArrayOutputStream();
          byte buf[] = new byte[1024];
          int r;
          while((r=pkIn.read(buf))>=0)
            bout.write(buf, 0, r);
          byte keyBytes[] = Util.decrypt(bout.toByteArray(), null);
          privKey = Util.getPrivateKeyFromBytes(keyBytes, 0);
          privKeyFile = null;
          idField.setText(DOConstants.ANONYMOUS_ID);
          wasCanceled = false;
          setVisible(false);
        } catch (Exception e) {
          JOptionPane.showMessageDialog(this, 
                                        "Error loading anonymous authentication: "+e,
                                        "Error", JOptionPane.ERROR_MESSAGE);
          return;
        }
      }
    } else if(src==loadKeyButton) {
      chooseKey();
    }
  }

  
  private class PassphrasePanel
  extends JPanel
  {
    private JPasswordField passField;
    
    PassphrasePanel() {
      setLayout(new GridBagLayout());
      setBorder(new javax.swing.border.EmptyBorder(10,10,10,10));
      passField = new JPasswordField("", 30);
      
      add(new JLabel("Enter the passphrase to decrypt your private key"),
          AwtUtil.getConstraints(0,0,1,1,1,1,false,false));
      add(passField, AwtUtil.getConstraints(0,1,1,1,1,1,false,false));
      
    }
    
    char[] getPassphrase() {
      return passField.getPassword();
    }
  }
  

}
